#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.linacs.elektaversa as versa
from opengate.tests import utility

if __name__ == "__main__":
    # paths
    paths = utility.get_default_test_paths(__file__, output_folder="test019_linac")

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.visu = False
    sim.visu_type = "vrml"
    sim.check_volumes_overlap = False
    sim.number_of_threads = 1
    sim.output_dir = paths.output
    sim.random_seed = 123456789
    sim.check_volumes_overlap = True
    sim.output_dir = paths.output
    sim.progress_bar = True

    # units
    nm = gate.g4_units.nm
    m = gate.g4_units.m
    mm = gate.g4_units.mm
    MeV = gate.g4_units.MeV

    # world
    world = sim.world
    world.size = [4 * m, 4 * m, 4 * m]
    world.material = "G4_AIR"

    # add a linac
    linac = versa.add_linac(sim, "versa")
    translation = [50 * mm, 12 * mm, 29 * mm]
    versa.translation_from_sad(sim, linac.name, translation, sad=1000)
    versa.rotation_around_user_point(
        sim, linac.name, "ZY", [38, 28], [224 * mm, -47 * mm, 456 * mm]
    )

    # add linac e- source
    source = versa.add_electron_source(sim, linac.name, 6.49 * MeV, 1 * mm, 0.5 * mm)
    source.n = 8e4 / sim.number_of_threads
    if sim.visu:
        source.n = 200

    # physics
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"
    sim.physics_manager.set_production_cut("world", "all", 1 * mm)
    versa.enable_brem_splitting(sim, linac.name, splitting_factor=10)

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "stats")
    stats.track_types_flag = True

    # add phase space
    plane = versa.add_phase_space_plane(sim, linac.name, linac.size[2] - 1 * nm)
    phsp = versa.add_phase_space_actor(sim, plane.name)
    phsp.output_filename = "phsp_versa.root"

    # start simulation
    sim.run()

    # print results
    print(stats)

    # compare root
    br = "versa_phsp_plane_phsp"
    print(paths.output_ref)
    root_ref = paths.output_ref / "phsp_versa_no_tr_no_rot.root"
    keys = ["KineticEnergy", "PrePositionLocal_X", "PrePositionLocal_Y"]
    tols = [0.1, 2.5, 2.5]
    is_ok = utility.compare_root3(
        root_ref,
        phsp.get_output_path(),
        br,
        br,
        keys,
        keys,
        tols,
        None,
        None,
        paths.output / "phsp_versa1.png",
        hits_tol=80,
    )

    root_ref = paths.output_ref / "phsp_versa_tr_no_rot.root"
    keys = ["KineticEnergy", "PrePositionLocal_X", "PrePositionLocal_Y"]
    tols = [0.1, 2.5, 2.5]
    is_ok = (
        utility.compare_root3(
            root_ref,
            phsp.get_output_path(),
            br,
            br,
            keys,
            keys,
            tols,
            None,
            None,
            paths.output / "phsp_versa2.png",
            hits_tol=80,
        )
        and is_ok
    )

    root_ref = paths.output_ref / "phsp_versa_tr_rot.root"
    keys = [
        "KineticEnergy",
        "PrePosition_X",
        "PrePosition_Y",
        "PrePositionLocal_X",
        "PrePositionLocal_Y",
    ]
    tols = [0.1, 2.5, 2.5, 2.5, 2.5]
    is_ok = (
        utility.compare_root3(
            root_ref,
            phsp.get_output_path(),
            br,
            br,
            keys,
            keys,
            tols,
            None,
            None,
            paths.output / "phsp_versa3.png",
            hits_tol=80,
        )
        and is_ok
    )
    # end
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from scipy.spatial.transform import Rotation
import opengate as gate
from opengate.tests import utility


if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "test_087")

    ref_path = paths.output_ref / "test087"
    mkm_lq_fpath = ref_path / "mkm_nirs_LQparameters_SURVIVAL.csv"
    #    df = pd.read_csv(mkm_lq_fpath)
    # create the simulation
    sim = gate.Simulation()

    # main options
    ui = sim.user_info
    ui.g4_verbose = False
    ui.g4_verbose_level = 1
    ui.visu = False
    ui.random_seed = 12345678910
    ui.number_of_threads = 8

    numPartSimTest = 4e2 / ui.number_of_threads
    numPartSimRef = 1e4

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    km = gate.g4_units.km
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    kBq = 1000 * Bq

    #  change world size
    world = sim.world
    world.size = [600 * cm, 500 * cm, 500 * cm]

    # waterbox
    phantom = sim.add_volume("Box", "phantom")
    phantom.size = [50 * mm, 10 * cm, 10 * cm]
    phantom.translation = [-25 * cm, 0, 0]
    phantom.material = "G4_WATER"
    phantom.color = [0, 0, 1, 1]

    test_material_name = "G4_WATER"
    phantom_off = sim.add_volume("Box", "phantom_off")
    phantom_off.mother = phantom.name
    phantom_off.size = [50 * mm, 10 * mm, 10 * mm]
    phantom_off.translation = [0 * mm, 0 * mm, 0 * mm]
    phantom_off.material = test_material_name
    phantom_off.color = [0, 0, 1, 1]

    sim.physics_manager.physics_list_name = "FTFP_INCLXX_EMZ"

    # default source for tests
    source = sim.add_source("GenericSource", "mysource")
    source.energy.mono = 1424 * MeV
    source.particle = "ion 6 12"
    source.position.type = "disc"
    source.position.rotation = Rotation.from_euler("y", 90, degrees=True).as_matrix()
    source.position.radius = 4 * mm
    source.position.translation = [0 * mm, 0, 0]
    source.direction.type = "momentum"
    source.direction.momentum = [-1, 0, 0]
    source.n = numPartSimTest

    size = [50, 1, 1]
    spacing = [1.0 * mm, 100.0 * mm, 100.0 * mm]

    doseActorName_IDD_d = "IDD_d"
    doseIDD = sim.add_actor("DoseActor", doseActorName_IDD_d)
    doseIDD.output_filename = paths.output / ("test087-" + doseActorName_IDD_d + ".mhd")
    #    print(f'actor: {paths.output / ("test087-" + doseActorName_IDD_d + ".mhd")}')
    doseIDD.attached_to = phantom_off.name
    doseIDD.size = size
    doseIDD.spacing = spacing
    doseIDD.hit_type = "random"
    doseIDD.dose.active = False

    RBE = "RBE"
    RBE_act = sim.add_actor("RBEActor", "RBE_act")
    RBE_act.output_filename = paths.output / ("test087-" + RBE + ".mhd")
    RBE_act.attached_to = phantom_off.name
    RBE_act.size = size
    RBE_act.spacing = spacing
    RBE_act.hit_type = "random"
    RBE_act.model = "mMKM"
    RBE_act.r_nucleus = 3.9
    #    RBE_act.model = "LEM1lda"

    RBE_act.energy_per_nucleon = False
    RBE_act.lookup_table_path = mkm_lq_fpath

    s = sim.add_actor("SimulationStatisticsActor", "stats")
    s.track_types_flag = True

    sim.run()

    ref_fpath = ref_path / "test087-ref_rbe.mhd"
    #    print(f"{doseIDD.dose.get_output_path()=}")

    is_ok = utility.assert_filtered_imagesprofile1D(
        ref_filter_filename1=doseIDD.edep.get_output_path(),
        ref_filename1=ref_fpath,
        filename2=paths.output / RBE_act.rbe.get_output_path(),
        tolerance=20,
        eval_quantity="RBE",
        #        plt_ylim=[0, 2],
    )

    ref_fpath = ref_path / "test087-ref_alpha_mix.mhd"

    fName = paths.output / RBE_act.alpha_mix.get_output_path()

    is_ok = is_ok and utility.assert_filtered_imagesprofile1D(
        ref_filter_filename1=doseIDD.edep.get_output_path(),
        ref_filename1=ref_fpath,
        filename2=paths.output / RBE_act.alpha_mix.get_output_path(),
        tolerance=20,
        eval_quantity="alpha",
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import uproot
import numpy as np
import opengate.tests.utility as utility


g_cm3 = gate.g4_units.g_cm3
mm = gate.g4_units.mm
cm = gate.g4_units.cm
m = gate.g4_units.m
Bq = gate.g4_units.Bq
MeV = gate.g4_units.MeV
sec = gate.g4_units.second
ns = gate.g4_units.nanosecond
deg = gate.g4_units.degree


def create_simu(material):

    paths = utility.get_default_test_paths(__file__, output_folder="test088")

    sim = gate.Simulation()
    sim.g4_verbose = False
    sim.visu = False
    sim.visu_type = "vrml"
    sim.random_engine = "MersenneTwister"
    sim.random_seed = 1234
    sim.output_dir = paths.output
    sim.number_of_threads = 1

    sim.world.size = [3 * m, 3 * m, 3 * m]
    sim.world.material = "G4_AIR"
    file = paths.data / "GateMaterials_Isotopes.db"
    sim.volume_manager.add_material_database(str(file))

    # waterbox
    waterbox = sim.add_volume("Box", "waterbox")
    waterbox.size = [40 * cm, 40 * cm, 40 * cm]
    waterbox.translation = [0 * cm, 0 * cm, 0 * cm]
    waterbox.material = material
    waterbox.color = [0, 1, 1, 1]

    # physics
    # sim.physics_manager.physics_list_name = "QGSP_BIC"
    sim.physics_manager.physics_list_name = "QGSP_BIC_HP"
    sim.physics_manager.set_production_cut("world", "all", 10 * mm)
    sim.physics_manager.set_production_cut("waterbox", "all", 0.01 * mm)

    # source
    source = sim.add_source("GenericSource", "mysource")
    source.particle = "gamma"
    source.energy.type = "gauss"
    source.energy.mono = 50 * MeV
    source.energy.sigma_gauss = 0.1 * MeV
    source.position.type = "disc"
    source.position.radius = 5 * cm
    source.position.translation = [0, 0, -100 * cm]
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    # source.n = 100000
    source.n = 10000

    # stats
    stats = sim.add_actor("SimulationStatisticsActor", "stats")
    stats.track_types_flag = True
    stats.output_filename = "stats.txt"

    # phase space actor
    phsp = sim.add_actor("PhaseSpaceActor", "PhaseSpace")
    phsp.attached_to = waterbox.name
    phsp.attributes = [
        "KineticEnergy",
    ]
    phsp.output_filename = "ps.root"
    phsp.steps_to_store = "exiting"
    f = sim.add_filter("ParticleFilter", "f")
    f.particle = "neutron"
    phsp.filters.append(f)

    # run
    sim.run(start_new_process=True)

    events = uproot.open(phsp.get_output_path_string())["PhaseSpace;1"]
    ekin = events["KineticEnergy"].array(library="np")

    return np.shape(ekin)[0]


if __name__ == "__main__":
    # Run a simulation without and with isotope
    # The simulation with isotope should output much more detected neutrons
    print("Run the simulation with normal water")
    neutron_water = create_simu("G4_WATER")
    print("Number of detected neutrons: " + str(neutron_water))
    print()
    print("Run the simulation with heavy water")
    neutron_water_h2 = create_simu("Water_H2")
    print("Number of detected neutrons: " + str(neutron_water_h2))
    print()
    print("Run the simulation with Deuterium")
    neutron_deuterium = create_simu("Deuterium")
    print("Number of detected neutrons: " + str(neutron_deuterium))
    print()
    # Run a simulation with a material with frequency < 1
    print("Run the simulation with BrainBNCTB1025mMB")
    neutron_brainBNCTB1025mMB = create_simu("BrainBNCTB1025mMB")
    print("Number of detected neutrons: " + str(neutron_brainBNCTB1025mMB))

    # Compute std (poisson distribution)
    std = np.sqrt(neutron_water)
    is_ok = neutron_water_h2 > (neutron_water + 2 * std)
    utility.print_test(
        is_ok,
        f"neutron_water_h2 {neutron_water_h2} > neutron_water + 2 * std {neutron_water + 2 * std}",
    )
    b = neutron_deuterium > (neutron_water + 2 * std)
    utility.print_test(
        b,
        f"neutron_deuterium {neutron_deuterium} > neutron_water + 2 * std {neutron_water + 2 * std}",
    )
    is_ok = is_ok and b

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
from opengate.sources.utility import set_source_energy_spectrum
import numpy as np
import gatetools

import matplotlib.pyplot as plt


def run_simulation(paths, interpolation: str = None):
    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.number_of_threads = 1
    sim.random_seed = 987654321
    sim.output_dir = paths.output

    # units
    mm = gate.g4_units.mm
    g_cm3 = gate.g4_units.g_cm3

    # materials
    sim.volume_manager.material_database.add_material_weights(
        "Vacuum", ["H"], [1], 1e-12 * g_cm3
    )

    world = sim.world
    world.size = [10 * mm, 10 * mm, 10 * mm]
    world.material = "Vacuum"

    source = add_source_energy_spectrum_histogram(sim, interpolation)

    # actors
    # stats = sim.add_actor("SimulationStatisticsActor", "Stats")

    phsp_actor = sim.add_actor("PhaseSpaceActor", "phsp_actor")
    phsp_actor.output_filename = (
        f"test010_energy_spectrum_histogram_{interpolation}.root"
    )
    phsp_actor.steps_to_store = "first"
    phsp_actor.attributes = [
        "KineticEnergy",
    ]

    # verbose
    sim.g4_commands_after_init.append("/tracking/verbose 0")

    # start simulation
    sim.run(start_new_process=True)

    # get info
    ekin = root_load_ekin(str(phsp_actor.get_output_path()))

    # test
    bin_edges = source.energy.spectrum_energy_bin_edges
    weights = source.energy.spectrum_weights

    src_data = np.array(weights)
    src_data = src_data / np.sum(src_data)

    sim_data, _ = np.histogram(ekin, bins=bin_edges, density=True)
    sim_data = sim_data / np.sum(sim_data)

    relerrs = (sim_data - src_data) / src_data
    oks = [
        utility.check_diff_abs(0, abs(relerrs[i]), tolerance=0.20, txt="relative error")
        for i in range(len(relerrs))
        if src_data[i] > 0.001
    ]
    is_ok = all(oks)

    plot(
        paths.output / f"test010_plot_energy_spectrum_histogram_{interpolation}.png",
        bin_edges,
        src_data,
        sim_data,
        relerrs,
    )

    utility.test_ok(is_ok)


def plot(output_file, bin_edges, src_data, sim_data, relerrs):
    fig, ax = plt.subplots(figsize=(8.5, 6))
    ax.set_xlabel("Energy (MeV)")
    ax.set_ylabel("Probability")
    ax.stairs(src_data, bin_edges, fill=True, label="input data")
    ax.stairs(sim_data, bin_edges, fill=True, alpha=0.6, label="simulation")

    ax.set_xscale("log")
    ax.set_yscale("log")

    ax2 = ax.twinx()
    ax2.set_ylabel("Relative error")
    ax2.set_ylim([-0.20, +0.20])
    ax2.plot(
        (bin_edges[:-1] + bin_edges[1:]) / 2,
        relerrs,
        label="relative error",
        color="C2",
    )

    ax.legend()

    fig.savefig(output_file, dpi=300)
    plt.close(fig)


def root_load_ekin(root_file: str):
    data_ref, keys_ref, _ = gatetools.phsp.load(root_file)

    index_ekin = keys_ref.index("KineticEnergy")
    ekin = [data_ref_i[index_ekin] for data_ref_i in data_ref]

    return ekin


def add_source_energy_spectrum_histogram(sim, interpolation: str = None):
    source = sim.add_source("GenericSource", "beam")
    source.particle = "e-"
    source.n = 2e6 / sim.number_of_threads
    source.position.type = "point"
    source.direction.type = "iso"

    set_source_energy_spectrum(source, "Lu177")  # After defining the particle!!
    source.energy.spectrum_histogram_interpolation = interpolation

    return source


if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "test010_generic_source_energy_spectrum", output_folder="test010"
    )

    run_simulation(paths, interpolation=None)
    # run_simulation(paths, interpolation="linear")  # relative error ok until last values



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from opengate.tests.utility import get_default_test_paths, test_ok, print_test
from opengate.utility import g4_units
from opengate.logger import INFO, DEBUG, RUN, NONE
from opengate.managers import Simulation


def rm_timing(s):
    return "\n".join(line for line in s.splitlines() if "Simulation: STOP." not in line)


if __name__ == "__main__":
    paths = get_default_test_paths(__file__, None, "test092")

    sim = Simulation()
    m = g4_units.m
    world = sim.world
    world.size = [3 * m, 3 * m, 3 * m]
    world.material = "G4_AIR"
    is_ok = True

    # no source, so a warning will be generated

    # add one actor for logger debug
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # default verbose = INFO
    sim.verbose_level = INFO
    print(f"Logger: {sim.verbose_level}")
    sim.log_sink = "str"
    sim.run(start_new_process=True)
    print(sim.log_output)

    # test
    ref1 = paths.output_ref / "test092_log_output1.txt"
    # create the reference
    # with open(ref1, "w") as f:
    #    f.write(sim.log_output)
    with open(ref1, "r") as f:
        ref = f.read()
    # remove the line with the time
    ref = rm_timing(ref)
    sim.log_output = rm_timing(sim.log_output)
    b = ref == sim.log_output and is_ok
    print_test(b, "Compare output with logger level INFO")

    # test with no log
    print()
    print("-" * 79)
    sim.verbose_level = NONE
    sim.log_sink = "str"
    print(f"Logger: {sim.verbose_level}")
    sim.run(start_new_process=True)
    print("->", sim.log_output, "<-")

    # test
    b = "" == sim.log_output and is_ok
    print_test(b, "Compare output with logger level INFO")

    # test with debug
    print()
    print("-" * 79)
    sim.verbose_level = DEBUG
    sim.log_sink = "str"
    print(f"Logger: {sim.verbose_level}")
    sim.run()  # start_new_process=True)
    print(sim.log_output)

    # test
    ref2 = paths.output_ref / "test092_log_output2.txt"
    # create the reference
    # with open(ref2, "w") as f:
    #    f.write(sim.log_output)
    with open(ref2, "r") as f:
        ref = f.read()
    ref = rm_timing(ref)
    sim.log_output = rm_timing(sim.log_output)
    b = ref == sim.log_output and is_ok
    print_test(b, "Compare output with logger level DEBUG")

    test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import test036_adder_depth_helpers as t036
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test036_adder_depth", "test036"
    )

    # create and run the simulation
    sim = t036.create_simulation("param", paths, "_par")

    # start simulation
    sim.run()

    # test the output
    is_ok = t036.test_output(sim, paths)

    utility.test_ok(is_ok)



from scipy.spatial.transform import Rotation
from box import Box
from opengate.utility import g4_units
import numpy as np


def create_material(simulation, name):
    gcm3 = g4_units.g_cm3
    simulation.volume_manager.material_database.add_material_nb_atoms(
        f"{name}_target_tungsten", ["W", "Re"], [9, 1], 19.4 * gcm3
    )
    simulation.volume_manager.material_database.add_material_weights(
        f"{name}_target_copper", ["Cu"], [1], 8.93 * gcm3
    )
    simulation.volume_manager.material_database.add_material_weights(
        f"{name}_colli", ["W", "Ni", "Fe"], [0.95, 0.0375, 0.0125], 18 * gcm3
    )
    simulation.volume_manager.material_database.add_material_weights(
        f"{name}_flat_filter", ["Cr", "Fe", "Ni"], [0.17, 0.75, 0.08], 7.9 * gcm3
    )
    simulation.volume_manager.material_database.add_material_weights(
        f"{name}_mylar", ["H", "C", "O"], [0.04196, 0.625016, 0.333024], 1.38 * gcm3
    )
    simulation.volume_manager.material_database.add_material_weights(
        f"{name}_carbon", ["C"], [1], 2.27 * gcm3
    )
    simulation.volume_manager.material_database.add_material_weights(
        f"{name}_aluminium", ["Al"], [1], 2.7 * gcm3
    )


def add_linac(sim, name="linac", sad=1000):
    # unit
    create_material(sim, name)

    # for debug : should be the same as create_material
    # sim.volume_manager.add_material_database('../contrib/elekta_synergy_materials.db')

    # check overlap
    sim.check_volumes_overlap = True

    # global box
    linac = add_empty_linac_box(sim, name, sad)

    # target
    add_target(sim, linac.name)

    # Primary collimator
    add_primary_collimator(sim, linac.name)

    # Flattening filter
    add_flattening_filter(sim, linac.name)

    # Ionizing chamber
    add_ionizing_chamber(sim, linac.name)

    # back_scatter_plate
    add_back_scatter_plate(sim, linac.name)

    # mirror
    add_mirror(sim, linac.name)

    # kill actor above the target
    kill_around_target(sim, linac.name)

    return linac


def add_empty_linac_box(sim, linac_name, sad=1000):
    # units
    m = g4_units.m
    mm = g4_units.mm
    linac = sim.add_volume("Box", linac_name)
    linac.material = "G4_AIR"
    linac.size = [1 * m, 1 * m, 0.6 * m]
    translation_linac_box = np.array([0 * mm, 0, sad - linac.size[2] / 2])
    linac.translation = translation_linac_box
    linac.color = [1, 1, 1, 0.8]
    return linac


def add_target(sim, linac_name):
    linac = sim.volume_manager.get_volume(linac_name)
    z_linac = linac.size[2]
    # unit
    mm = g4_units.mm
    nm = g4_units.nm

    # colors
    red = [1, 0.2, 0.2, 0.8]
    green = [0, 1, 0, 0.2]

    # material
    target_material = f"{linac_name}_target_tungsten"
    copper = f"{linac_name}_target_copper"

    # target
    target_support = sim.add_volume("Tubs", f"{linac_name}_target_support")
    target_support.mother = linac_name
    target_support.material = "G4_AIR"
    target_support.rmin = 0
    target_support.rmax = 15 * mm
    target_support.dz = 11 * mm / 2.0
    target_support.translation = [0, 0, z_linac / 2 - target_support.dz - 1 * nm]
    target_support.color = [0, 1, 0, 1]

    target = sim.add_volume("Tubs", f"{linac_name}_target")
    target.mother = target_support.name
    target.material = target_material
    target.rmin = 0
    target.rmax = 2.7 * mm
    target.dz = 1 * mm / 2.0
    target.translation = [0, 0, 5 * mm]
    target.color = red

    target_support_top = sim.add_volume("Tubs", f"{linac_name}_target_support_top")
    target_support_top.mother = target_support.name
    target_support_top.material = copper
    target_support_top.rmin = 2.7 * mm
    target_support_top.rmax = 15 * mm
    target_support_top.dz = 1 * mm / 2.0
    target_support_top.translation = [0, 0, 5 * mm]
    target_support_top.color = green

    target_support_bottom = sim.add_volume(
        "Tubs", f"{linac_name}_target_support_bottom"
    )
    target_support_bottom.mother = target_support.name
    target_support_bottom.material = copper
    target_support_bottom.rmin = 0
    target_support_bottom.rmax = 15 * mm
    target_support_bottom.dz = 10 * mm / 2.0
    target_support_bottom.translation = [0, 0, -0.5 * mm]
    target_support_bottom.color = green


def kill_around_target(sim, linac_name):
    mm = g4_units.mm
    nm = g4_units.nm
    linac = sim.volume_manager.get_volume(linac_name)
    target = sim.volume_manager.get_volume(f"{linac_name}_target")
    z_linac = linac.size[2]

    # above the target
    target_above = sim.add_volume("Tubs", f"{target.name}_kill_volume1")
    target_above.mother = linac_name
    target_above.material = "G4_AIR"
    target_above.rmin = 0
    target_above.rmax = 15 * mm
    target_above.dz = 0.5 * nm
    target_above.translation = [0, 0, z_linac / 2 - 0.5 * nm]
    target_above.color = [1, 0, 0, 1]

    # around the target
    target_around = sim.add_volume("Tubs", f"{target.name}_kill_volume2")
    target_around.mother = linac_name
    target_around.material = "G4_AIR"
    target_around.rmin = 15.1 * mm
    target_around.rmax = target_around.rmin + 1 * nm
    target_around.dz = 5.5 * mm
    target_around.translation = [0, 0, z_linac / 2 - target_around.dz - 1 * nm]
    target_around.color = [1, 0, 0, 1]

    # psycho killer
    killer = sim.add_actor("KillActor", f"{target.name}_kill")
    killer.attached_to = [target_above.name, target_around.name]


def add_primary_collimator(sim, linac_name):
    mm = g4_units.mm
    deg = g4_units.deg
    blue = [0.5, 0.5, 1, 0.8]
    linac = sim.volume_manager.get_volume(linac_name)
    z_linac = linac.size[2]
    primary_collimator = sim.add_volume("Cons", f"{linac_name}_primary_collimator")
    primary_collimator.mother = linac.name
    primary_collimator.material = f"{linac_name}_colli"
    primary_collimator.rmin1 = 31.45 * mm
    primary_collimator.rmax1 = 82 * mm
    primary_collimator.rmin2 = 6.45 * mm
    primary_collimator.rmax2 = 82 * mm
    primary_collimator.dz = 101 * mm / 2.0
    primary_collimator.sphi = 0
    primary_collimator.dphi = 360 * deg
    primary_collimator.translation = [0, 0, z_linac / 2 - 65.5 * mm]
    primary_collimator.color = blue


def add_flattening_filter(sim, linac_name):
    # unit
    mm = g4_units.mm
    deg = g4_units.deg

    # colors
    yellow = [0, 0.7, 0.7, 0.8]
    linac = sim.volume_manager.get_volume(linac_name)
    z_linac = linac.size[2]

    # bounding cylinder
    flattening_filter = sim.add_volume("Tubs", f"{linac_name}_flattening_filter")
    flattening_filter.mother = linac_name
    flattening_filter.material = "G4_AIR"
    flattening_filter.rmin = 0
    flattening_filter.rmax = 40 * mm
    flattening_filter.dz = 24.1 * mm / 2
    flattening_filter.translation = [0, 0, z_linac / 2 - 146.05 * mm]
    flattening_filter.color = [0, 0, 0, 0]  # invisible

    # create all cones
    def add_cone(sim, p):
        c = sim.add_volume("Cons", f"{linac_name}_flattening_filter_cone_{p.name}")
        c.mother = f"{linac_name}_flattening_filter"
        c.material = f"{linac_name}_flat_filter"
        c.rmin1 = 0
        c.rmax1 = p.rmax1
        c.rmin2 = 0
        c.rmax2 = p.rmax2
        c.dz = p.dz
        c.sphi = 0
        c.dphi = 360 * deg
        c.translation = [0, 0, p.tr]
        c.color = yellow

    cones = [
        [0.001, 5.45, 3.40, 10.35],
        [5.45, 9, 2.7, 7.3],
        [9, 14.5, 4.9, 3.5],
        [14.5, 22.5, 5.5, -1.7],
        [22.5, 32.5, 5.6, -7.25],
        [38.5, 38.5, 2, -11.05],
    ]  ## FIXME check 32.5 ?
    i = 0
    for c in cones:
        cone = Box()
        cone.name = i
        cone.rmax2 = c[0] * mm
        cone.rmax1 = c[1] * mm
        cone.dz = c[2] * mm / 2  # /2 to keep same values than Gate (where dz was /2)
        cone.tr = c[3] * mm
        add_cone(sim, cone)
        i = i + 1


def add_ionizing_chamber(sim, linac_name):
    # unit
    mm = g4_units.mm

    # main cylinder
    linac = sim.volume_manager.get_volume(linac_name)
    z_linac = linac.size[2]
    ionizing_chamber = sim.add_volume("Tubs", f"{linac_name}_ionizing_chamber")
    ionizing_chamber.mother = linac_name
    ionizing_chamber.material = "G4_AIR"
    ionizing_chamber.rmin = 0
    ionizing_chamber.rmax = 45 * mm
    ionizing_chamber.dz = 9.28 * mm / 2
    ionizing_chamber.translation = [0, 0, z_linac / 2 - 169 * mm]
    ionizing_chamber.color = [0, 0, 0, 0]

    # layers
    def add_layer(sim, p):
        l = sim.add_volume("Tubs", f"{linac_name}_ionizing_chamber_mylar_layer_{p.i}")
        l.mother = f"{linac_name}_ionizing_chamber"
        l.material = f"{linac_name}_mylar"
        l.rmin = 0
        l.rmax = 45 * mm
        l.dz = 0.012 * mm / 2
        l.translation = [0, 0, p.tr1]

        l = sim.add_volume("Tubs", f"{linac_name}_ionizing_chamber_carbon_layer_{p.i}")
        l.mother = f"{linac_name}_ionizing_chamber"
        l.material = f"{linac_name}_carbon"
        l.rmin = 0
        l.rmax = 45 * mm
        l.dz = 0.000150 * mm / 2
        l.translation = [0, 0, p.tr2]

    layers = [
        [-2.634, -2.627925],
        [-0.434, -0.427925],
        [0.566, 0.572075],
        [1.566, 1.572075],
        [2.566, 2.572075],
        [3.566, 3.572075],
    ]
    i = 1
    for l in layers:
        ll = Box()
        ll.i = i
        ll.tr1 = l[0] * mm
        ll.tr2 = l[1] * mm
        add_layer(sim, ll)
        i = i + 1


def add_back_scatter_plate(sim, linac_name):
    # back_scatter_plate
    mm = g4_units.mm
    red = [1, 0.7, 0.7, 0.8]
    linac = sim.volume_manager.get_volume(linac_name)
    z_linac = linac.size[2]
    mm = g4_units.mm
    bsp = sim.add_volume("Box", f"{linac_name}_back_scatter_plate")
    bsp.mother = linac_name
    bsp.material = f"{linac_name}_aluminium"
    bsp.size = [116 * mm, 84 * mm, 3 * mm]
    bsp.translation = [0, 0, z_linac / 2 - 183 * mm]
    bsp.color = red


def add_mirror(sim, linac_name):
    # unit
    mm = g4_units.mm
    blue = [0, 0, 1, 0.8]

    # main box
    linac = sim.volume_manager.get_volume(linac_name)
    z_linac = linac.size[2]
    m = sim.add_volume("Box", f"{linac_name}_mirror")
    m.mother = linac_name
    m.material = "G4_AIR"
    m.size = [137 * mm, 137 * mm, 1.5 * mm]
    m.translation = [0, 0, z_linac / 2 - 225 * mm]
    rot = Rotation.from_euler("x", 37.5, degrees=True)
    m.rotation = rot.as_matrix()

    # mylar
    l = sim.add_volume("Box", f"{linac_name}_mirror_mylar_layer")
    l.mother = m.name
    l.material = f"{linac_name}_mylar"
    l.size = [110 * mm, 110 * mm, 0.0012 * mm]
    l.translation = [0, 0, 0.15 * mm]
    l.color = blue

    # alu
    l = sim.add_volume("Box", f"{linac_name}_mirror_alu_layer")
    l.mother = m.name
    l.material = f"{linac_name}_aluminium"
    l.size = [110 * mm, 110 * mm, 0.0003 * mm]
    l.translation = [0, 0, -0.6 * mm]
    l.color = blue


def enable_brem_splitting(sim, linac_name, splitting_factor):
    # FIXME: consider using new biasing mechanism instead of G4 commands
    # create a region
    linac = sim.volume_manager.get_volume(linac_name)
    region_linac = sim.physics_manager.add_region(name=f"{linac.name}_region")
    region_linac.associate_volume(linac)
    # set the brem splitting
    s = f"/process/em/setSecBiasing eBrem {region_linac.name} {splitting_factor} 50 MeV"
    sim.g4_commands_after_init.append(s)


def add_electron_source(sim, linac_name, rotation_matrix):
    MeV = g4_units.MeV
    mm = g4_units.mm
    nm = g4_units.nm
    target = sim.volume_manager.get_volume(f"{linac_name}_target")
    source = sim.add_source("GenericSource", f"{linac_name}_e-_source")
    source.particle = "e-"
    source.attached_to = f"{linac_name}_target"
    source.energy.type = "gauss"
    source.energy.mono = 6.7 * MeV
    source.energy.sigma_gauss = 0.077 * MeV
    source.position.type = "disc"
    source.position.radius = 2 * mm  # FIXME not really similar to GATE need sigma etc
    source.position.translation = [0, 0, target.dz - 1 * nm]
    source.direction.type = "momentum"
    source.n = 10
    # consider linac rotation
    dir = np.dot(rotation_matrix, np.array([0, 0, -1]))
    source.direction.momentum = dir
    return source


def add_phase_space_plane(sim, linac_name, src_phsp_distance):
    mm = g4_units.mm
    m = g4_units.m
    nm = g4_units.nm
    plane = sim.add_volume("Tubs", f"{linac_name}_phsp_plane")
    plane.mother = linac_name
    plane.material = "G4_AIR"
    plane.rmin = 0
    plane.rmax = 0.5 * m / 2 - 0.1 * mm
    plane.dz = 1 * nm  # half height
    linac = sim.volume_manager.get_volume(linac_name)
    z_linac = linac.size[2]
    plane.translation = [0 * mm, 0 * mm, +z_linac / 2 - src_phsp_distance]
    plane.color = [1, 0, 0, 1]  # red
    return plane


def add_phase_space_actor(sim, plane_name):
    phsp = sim.add_actor("PhaseSpaceActor", f"{plane_name}_phsp")
    phsp.attached_to = plane_name
    phsp.attributes = [
        "KineticEnergy",
        "Weight",
        "PrePosition",
        "PrePositionLocal",
        "PreDirection",
        "PreDirectionLocal",
        "PDGCode",
    ]
    return phsp



from box import Box
import pathlib
import numpy as np
import json
import re
import icrp107_database

from ..utility import g4_units
from ..exception import fatal

gate_source_path = pathlib.Path(__file__).parent.resolve()

# http://www.lnhb.fr/nuclear-data/module-lara/
all_beta_plus_radionuclides = [
    "F18",
    "Ga68",
    "Zr89",
    "Na22",
    "C11",
    "N13",
    "O15",
    "Rb82",
]

icrp107_emissions = [
    "alpha",
    "beta-",
    "beta+",
    "gamma",
    "X",
    "neutron",
    "auger",
    "IE",
    "alpha recoil",
    "anihilation",
    "fission",
    "betaD",
    "b-spectra",  # beta spectras, both beta+ and beta-
]

DEFAULT_DATABASE = "icrp107"
DEFAULT_SPECTRUM_TYPE = "gamma"


def get_spectrum(
    rad_name: str, spectrum_type=DEFAULT_SPECTRUM_TYPE, database=DEFAULT_DATABASE
) -> Box:
    """
    Retrieve the spectrum of a given radionuclide from the specified database.

    Parameters
    ----------
    rad_name : str
        The name of the radionuclide in Gate format (e.g. "Tc99m", "Lu177").

    spectrum_type : str, optional
        The type of spectrum to retrieve. Default is "gamma". Must be one of
        "gamma", "beta+", "beta-", or "e+". icrp107 allows also one of
        "alpha", "X", "neutron", "auger", "IE", "alpha recoil", "anihilation", "fission", "betaD", "b-spectra".
        In the case of beta spectras, make use of the
        :py:func:`set_source_energy_spectrum` function instead.

    database : str, optional
        The database to retrieve the spectrum from. Default is "icrp107".
        If not "icrp107", we use radar data for gammas and beta-, and LNHB data for beta+.

    Returns
    -------
    Box
        A Box object containing the spectrum data.

    Raises
    ------
    fatal
        If the specified database does not contain the requested spectrum type.
    """
    if database == "icrp107":
        return __get_icrp107_spectrum(rad_name, spectrum_type)
    else:
        if spectrum_type == "beta+" or spectrum_type == "e+":
            return __read_beta_plus_spectra(rad_name)
        elif spectrum_type == "beta-" or spectrum_type == "e-":
            return __get_rad_beta_spectrum(rad_name)
        elif spectrum_type == "gamma":
            return __get_rad_gamma_spectrum(rad_name)
        else:  # FIXME use icrp107 for missing spectrum types
            fatal(f"databse {database} doesn't contain spectrum type {spectrum_type}")


def set_source_energy_spectrum(source, rad: str, database=DEFAULT_DATABASE) -> None:
    """
    Set the energy spectrum of a source.

    Parameters
    ----------
    source : a source
        The source to set the energy spectrum for
    rad : str
        The name of the radionuclide to use
    database : str, optional
        The database to use. The default is "icrp107".

    Notes
    -----
    The source particle must be set before calling this function.
    If the source particle is "beta-/e-" or "beta+/e+", the function will use the
    "b-spectra" spectrum from the database data.
    Otherwise, the function will use the discrete spectrum for the given particle.

    """
    if (
        source.particle == "beta-"
        or source.particle == "e-"
        or source.particle == "beta+"
        or source.particle == "e+"
    ):
        rad_spectrum = get_spectrum(rad, "b-spectra", database)
        source.energy.type = "spectrum_histogram"
        source.energy.spectrum_weights = rad_spectrum.weights[:-1]
        source.energy.spectrum_energy_bin_edges = rad_spectrum.energies
    else:
        rad_spectrum = get_spectrum(rad, source.particle, database)
        source.energy.type = "spectrum_discrete"
        source.energy.spectrum_weights = rad_spectrum.weights
        source.energy.spectrum_energies = rad_spectrum.energies


def __gate_radname_to_icrp107(rad_name: str) -> str:
    """
    Convert a radionuclide name from GATE to ICRP 107 format.

    GATE format is for example "F18" and ICRP 107 format is "F-18"

    Parameters
    ----------
    rad_name : str
        radionuclide name in GATE format

    Returns
    -------
    str
        radionuclide name in ICRP 107 format
    """
    excited = rad_name[-1] == "m"  # Handle Tc-99m
    at_num = re.findall(r"\d+", rad_name)[0]  # Extract atomic number
    name = rad_name[:-1] if excited else rad_name  # Remove final m
    elem = name.replace(at_num, "").replace("-", "")  # Find element code
    elem = elem[0].upper() + elem[1:]  # Fix
    return f'{elem}-{at_num}{"m" if excited else ""}'


icrp107_time_units = {
    "ms": g4_units.millisecond,
    "s": g4_units.second,
    "m": g4_units.minute,
    "h": g4_units.hour,
    "d": g4_units.day,
    "y": g4_units.year,
}


def __convert_icrp107_time_unit(icrp_time_unit: str) -> float:
    """
    Convert an ICRP 107 time unit to its corresponding value in GATE units.

    Parameters
    ----------
    icrp_time_unit : str
        The time unit from ICRP 107. Valid options are "ms" (milliseconds),
        "s" (seconds), "m" (minutes), "h" (hours), "d" (days), and "y" (years).

    Returns
    -------
    float
        The equivalent time value in GATE units.

    Raises
    ------
    Exception
        If the provided time unit is not recognized.
    """
    if icrp_time_unit in icrp107_time_units:
        return icrp107_time_units[icrp_time_unit]
    else:
        fatal(f"unit {icrp_time_unit} not recognized")


def __read_beta_plus_spectra(rad_name):
    """
    read the file downloaded from LNHB
    there are 15 lines-long header to skip
    first column is E(keV)
    second column is dNtot/dE b+
    WARNING : bins width is not uniform (need to scale for density)
    """
    if rad_name not in all_beta_plus_radionuclides:
        # FIXME use icrp107 for missing isotopes
        fatal(f"rad_name {rad_name} not in {all_beta_plus_radionuclides}")

    filename = (
        f"{gate_source_path}/beta_plus_spectra/{rad_name}/beta+_{rad_name}_tot.bs"
    )
    data = np.genfromtxt(filename, usecols=(0, 1), skip_header=15, dtype=float)
    # FIXME convert to MeV before return
    return data


def compute_bins_density(bins):
    """
    Given a list of (energy) bins center, compute the width of each bin.
    """
    lower = np.roll(bins, 1)
    lower[0] = 0
    upper = bins
    dx = upper - lower
    return dx


def get_rad_yield(rad_name):
    if rad_name not in all_beta_plus_radionuclides:
        return 1.0
    data = __read_beta_plus_spectra(rad_name)
    ene = data[:, 0] / 1000  # convert from KeV to MeV
    proba = data[:, 1]
    _, total = compute_cdf_and_total_yield(proba, ene)
    total = total * 1000  # (because was in MeV)
    return total


def compute_cdf_and_total_yield(data, bins):
    """
    Compute the CDF (Cumulative Density Function) of a list of non-uniform energy bins
    with associated probability.
    Also return the total probability.
    """
    dx = compute_bins_density(bins)
    p = data * dx
    total = p.sum()
    cdf = np.cumsum(p) / total
    return cdf, total


def __get_icrp107_spectrum(rad_name: str, spectrum_type=DEFAULT_SPECTRUM_TYPE) -> Box:
    """
    Get the spectrum of a given radionuclide according to ICRP107 recommendations.

    Parameters
    ----------
    rad_name : str
        The name of the radionuclide in Gate format, e.g. "Tc99m", "Lu177"

    spectrum_type : str
        The type of spectrum to retrieve. Must be one of "gamma", "beta-", "beta+", "alpha", "X", "neutron", "auger", "IE", "alpha recoil", "anihilation", "fission", "betaD", "b-spectra"

    Returns
    -------
    Box
        A Box with two keys: "energies" and "weights". The first contains the energy of each emission, the second contains the weight of each emission (summing to 1).

    Raises
    ------
    fatal
        If the radionuclide or spectrum type is not valid.
    """
    rad = __gate_radname_to_icrp107(rad_name)
    icrp107_data = icrp107_database.get_icrp107_spectrum(rad, spectrum_type)

    # convert to Box with unit
    gate_data = {}
    gate_data["energies"] = np.array(
        [v * g4_units.MeV for v in icrp107_data["energies"]]
    )
    gate_data["weights"] = np.array([v for v in icrp107_data["weights"]])
    gate_data["half_life"] = icrp107_data["half_life"] * __convert_icrp107_time_unit(
        icrp107_data["time_unit"]
    )
    return Box(gate_data)


def __get_rad_gamma_spectrum(rad):
    path = gate_source_path.parent / "data" / "rad_gamma_spectrum.json"
    with open(path, "r") as f:
        data = json.load(f)

    # consider lower case
    data = {key.lower(): value for key, value in data.items()}
    rad = rad.lower()

    if rad not in data:
        # FIXME use icrp107 for missing isotopes
        fatal(f"get_rad_gamma_spectrum: {path} does not contain data for ion {rad}")

    # select data for specific ion
    data = Box(data[rad])

    data.energies = np.array(data.energies) * g4_units.MeV
    data.weights = np.array(data.weights)

    return data


def __get_rad_beta_spectrum(rad: str):
    path = gate_source_path.parent / "data" / "rad_beta_spectrum.json"
    with open(path, "r") as f:
        data = json.load(f)

    if rad not in data:
        # FIXME use icrp107 for missing isotopes
        fatal(f"get_rad_beta_spectrum: {path} does not contain data for ion {rad}")

    # select data for specific ion
    data = Box(data[rad])

    bin_edges = data.energy_bin_edges
    n = len(bin_edges) - 1
    energies = [(bin_edges[i] + bin_edges[i + 1]) / 2 for i in range(n)]

    data.energy_bin_edges = np.array(data.energy_bin_edges) * g4_units.MeV
    data.weights = np.array(data.weights)
    data.energies = np.array(energies)

    return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.spect.ge_discovery_nm670 as nm670
from opengate.contrib.spect.spect_helpers import get_default_energy_windows
import opengate.contrib.phantoms.nemaiec as nemaiec
from opengate.image import get_translation_to_isocenter
from opengate.sources.utility import set_source_energy_spectrum
from pathlib import Path
import numpy as np
import opengate_core as g4


def check_process_user_hook(simulation_engine):
    p_name = "gamma"
    g4_particle_table = g4.G4ParticleTable.GetParticleTable()
    particle = g4_particle_table.FindParticle(particle_name=p_name)
    if particle is None:
        raise Exception(f"Something went wrong. Could not find particle {p_name}.")
    pm = particle.GetProcessManager()
    process_list = pm.GetProcessList()
    for i in range(process_list.size()):
        processName = str(process_list[i].GetProcessName())
        print("Checking process", processName)


def create_simulation_test085(
    sim,
    paths,
    simu_name,
    ac=1e5,
    angle_tolerance=None,
    use_spect_head=False,
    use_spect_arf=False,
    use_phsp=False,
):
    # main options
    sim.visu_type = "qt"
    # sim.visu_type = "vrml"
    sim.random_seed = "auto"
    sim.progress_bar = True
    sim.output_dir = paths.output
    sim.store_json_archive = True
    sim.store_input_files = False
    sim.json_archive_filename = f"simu_{simu_name}.json"
    sim.random_seed = 123234
    data_folder = Path(paths.data) / "test085"

    # units
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    m = gate.g4_units.m
    Bq = gate.g4_units.Bq
    cm3 = gate.g4_units.cm3
    BqmL = Bq / cm3
    deg = gate.g4_units.deg

    # options
    activity = ac * BqmL / sim.number_of_threads
    radius = 28 * cm

    # visu
    if sim.visu:
        sim.number_of_threads = 1
        activity = 50 * BqmL / sim.number_of_threads

    # world
    world = sim.world
    world.size = [2 * m, 2 * m, 2 * m]
    world.material = "G4_Galactic"

    # GeneralProcess must *NOT* be true (it is by default)
    s = f"/process/em/UseGeneralProcess false"
    sim.g4_commands_before_init.append(s)

    # check option
    if use_spect_head and use_spect_arf:
        raise Exception("Cannot use both spect heads and ARFs.")

    # set the two spect heads
    spacing = [2.21 * mm * 2, 2.21 * mm * 2]
    size = [128, 128]
    heads = []
    actors = []
    if use_spect_arf:
        actors, heads = add_spect_arf(
            sim, data_folder, simu_name, radius, size, spacing
        )
    if use_spect_head:
        actors, heads = add_spect_heads(sim, simu_name, radius)

    # add phsp to check E spectra
    if use_phsp:
        actors, heads = add_phsp(
            sim, simu_name, radius, size, spacing, use_parallel_world=use_spect_head
        )
        # IEC voxelization
        # voxelize_iec_phantom -o data/iec_1mm.mhd --spacing 1 --output_source data/iec_1mm_activity.mhd -a 1 1 1 1 1 1
        # voxelize_iec_phantom -o data/iec_4.42mm.mhd --spacing 4.42 --output_source data/iec_4.42mm_activity.mhd -a 1 1 1 1 1 1
        # voxelize_iec_phantom -o data/iec_4mm.mhd --spacing 4 --output_source data/iec_4mm_activity.mhd -a 1 1 1 1 1 1

    # phantom
    if not sim.visu:
        iec_vox_filename = data_folder / "iec_4mm.mhd"
        iec_label_filename = data_folder / "iec_4mm_labels.json"
        db_filename = data_folder / "iec_4mm.db"
        vox = sim.add_volume("ImageVolume", "phantom")
        vox.image = iec_vox_filename
        vox.read_label_to_material(iec_label_filename)
        vox.translation = get_translation_to_isocenter(vox.image)
        sim.volume_manager.add_material_database(str(db_filename))
    else:
        phantom = nemaiec.add_iec_phantom(sim, name="phantom")

    # physics
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"
    sim.physics_manager.set_production_cut("world", "all", 100 * m)
    sim.physics_manager.set_production_cut("phantom", "all", 1 * mm)
    # sim.physics_manager.set_production_cut("phantom", "gamma", 0.01 * mm)
    sim.user_hook_after_init = check_process_user_hook

    # add iec voxelized source
    iec_source_filename = data_folder / "iec_4mm_activity.mhd"
    source = sim.add_source("VoxelSource", "src")
    source.image = iec_source_filename
    source.position.translation = [0, 35 * mm, 0]
    set_source_energy_spectrum(source, "tc99m", "radar")
    source.particle = "gamma"
    _, volumes = nemaiec.get_default_sphere_centers_and_volumes()
    source.activity = activity * np.array(volumes).sum()
    print(f"Total activity is {source.activity / Bq}")

    # add a stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "stats")
    stats.track_types_flag = True
    stats.output_filename = f"stats_{simu_name}.txt"

    # set the gantry orientation
    starting_angle_deg = 10
    if len(heads) == 2:
        nm670.set_default_orientation(heads[0], "lehr")
        nm670.set_default_orientation(heads[1], "lehr")
        nm670.rotate_gantry(heads[0], radius, starting_angle_deg, 0, 1)
        nm670.rotate_gantry(heads[1], radius, starting_angle_deg + 180, 0, 1)

    return source, actors


def add_spect_arf(sim, data_folder, simu_name, radius, size, spacing):
    pth = data_folder / "arf_034_nm670_tc99m_v2.pth"
    det_plane1, arf1 = nm670.add_arf_detector_OLD(
        sim, radius, 0, size, spacing, "lehr", "spect", 1, pth
    )
    det_plane2, arf2 = nm670.add_arf_detector_OLD(
        sim, radius, 180, size, spacing, "lehr", "spect", 2, pth
    )
    det_planes = [det_plane1, det_plane2]

    arf1.output_filename = f"projection_1_{simu_name}.mhd"
    arf2.output_filename = f"projection_2_{simu_name}.mhd"
    arfs = [arf1, arf2]
    return arfs, det_planes


def add_spect_heads(sim, simu_name, radius):
    mm = gate.g4_units.mm
    heads, crystals = nm670.add_spect_two_heads(
        sim, "spect", "lehr", debug=sim.visu, radius=radius
    )
    channels = get_default_energy_windows("tc99m")
    digit1 = nm670.add_digitizer(sim, crystals[0].name, "digit1", channels=channels)
    digit2 = nm670.add_digitizer(sim, crystals[1].name, "digit2", channels=channels)

    # we need the weights for the digitizer
    hits1 = digit1.find_module("hits")
    hits1.attributes.append("Weight")
    hits1.attributes.append("TrackID")

    hits2 = digit2.find_module("hits")
    hits2.attributes.append("Weight")
    hits2.attributes.append("TrackID")

    proj1 = digit1.find_module("projection")
    proj1.output_filename = f"projection_1_{simu_name}.mhd"
    proj2 = digit2.find_module("projection")
    proj2.output_filename = f"projection_2_{simu_name}.mhd"
    proj1.squared_counts.active = True
    proj2.squared_counts.active = True
    projs = [proj1, proj2]

    # sim.physics_manager.set_production_cut(crystals[0].name, "all", 2 * mm)
    # sim.physics_manager.set_production_cut(crystals[1].name, "all", 2 * mm)
    sim.physics_manager.set_production_cut("spect_1", "all", 2 * mm)
    sim.physics_manager.set_production_cut("spect_2", "all", 2 * mm)

    return projs, heads


def add_phsp(sim, simu_name, radius, size, spacing, use_parallel_world, sph_rad=None):

    phsp_sphere = sim.add_volume("Sphere", "phsp_sphere")
    if sph_rad is None:
        phsp_sphere.rmax = 30 * gate.g4_units.cm
    else:
        phsp_sphere.rmax = sph_rad
    phsp_sphere.rmin = phsp_sphere.rmax - 2 * gate.g4_units.mm
    phsp_sphere.color = [1, 0, 0, 1]

    phsp1 = sim.add_actor("PhaseSpaceActor", "phsp_sphere")
    phsp1.attached_to = phsp_sphere
    phsp1.attributes = ["KineticEnergy", "PrePositionLocal", "Weight"]
    phsp1.output_filename = f"phsp_sphere_{simu_name}.root"

    # gamma only
    fe = sim.add_filter("ParticleFilter", "fe")
    fe.particle = "gamma"
    fe.policy = "accept"
    phsp1.filters.append(fe)

    phsps = [phsp1]
    planes = [phsp_sphere]

    return phsps, planes



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test073_helpers import *
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, gate_folder="", output_folder="test073"
    )

    # create the simulation
    sim = gate.Simulation()
    head, stats, source = test073_setup_sim(sim, "intevo", collimator_type="lehr")
    sim.random_seed = 2974516
    sim.output_dir = paths.output

    # digit
    crystal = sim.volume_manager.get_volume(f"{head.name}_crystal")
    digit = intevo.add_digitizer(sim, crystal.name, "digit_tc99m")

    # add a channel 'spectrum' (which is not by default because not compatible with ARF)
    keV = gate.g4_units.keV
    c = {"name": f"spectrum", "min": 3 * keV, "max": 160 * keV}
    ew = digit.find_module("energy_window")
    ew.output_filename = "output_tc99m.root"
    ew.root_output.write_to_disk = True
    ew.channels.append(c)

    # output
    stats.output_filename = "stats_intevo_tc99m.txt"

    # source
    Bq = gate.g4_units.Bq
    set_source_energy_spectrum(source, "Tc99m")
    source.activity = 2e7 * Bq / sim.number_of_threads
    if sim.visu:
        source.activity = 1e3 * Bq

    # start simulation
    sim.run()

    # stat
    print(stats)

    # compare stats
    ref_folder = paths.output_ref
    is_ok = compare_stats(sim, ref_folder / "stats_intevo_tc99m.txt")

    # compare root
    fr = ref_folder / "output_intevo_tc99m.root"
    is_ok = (
        compare_root_spectrum(
            fr, ew.get_output_path(), paths.output / "test073_intevo_tc99m.png"
        )
        and is_ok
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from box import Box
from opengate.contrib.dose.doserate import create_simulation
from opengate.tests import utility


if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "", output_folder="test035")
    dr_data = paths.data / "dose_rate_data"

    # set param
    gcm3 = gate.g4_units.g_cm3
    param = Box()
    param.ct_image = str(dr_data / "29_CT_5mm_crop.mhd")
    param.table_mat = str(dr_data / "Schneider2000MaterialsTable.txt")
    param.table_density = str(dr_data / "Schneider2000DensitiesTable.txt")
    param.activity_image = str(dr_data / "activity_test_crop_4mm.mhd")
    param.radionuclide = "Lu177"
    param.activity_bq = 1e6
    param.number_of_threads = 1
    param.visu = False
    param.verbose = True
    param.density_tolerance_gcm3 = 0.05
    param.output_folder = str(paths.output)

    # Create the simu
    # Note that the returned sim object can be modified to change source or cuts or whatever other parameters
    sim = create_simulation(param)

    # stats
    stats = sim.get_actor("Stats")
    stats.output_filename = "stats035.txt"

    # Change source to alpha to get quick high local dose
    source = sim.get_source_user_info("vox")
    source.particle = "alpha"
    MeV = gate.g4_units.MeV
    source.energy.mono = 1 * MeV

    print("Phys list cuts:")
    print(sim.physics_manager.dump_production_cuts())

    # run
    sim.run(start_new_process=True)

    # print results
    print()
    print(stats)
    stats_ref = utility.read_stat_file(paths.output_ref / "stats.txt")
    is_ok = utility.assert_stats(stats, stats_ref, 0.10)

    # dose comparison
    print()
    gate.exception.warning(f"Check dose")
    h = sim.get_actor("dose")
    print(h)
    is_ok = (
        utility.assert_images(
            paths.output_ref / "edep.mhd",
            h.edep.get_output_path(),
            stats,
            tolerance=15,
            ignore_value_data2=0,
        )
        and is_ok
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from opengate.tests.utility import *
from opengate.sources.phidsources import *
from opengate.utility import g4_units
import numpy as np
import math


def create_ion_gamma_simulation(sim, paths, z, a):
    # find ion name and direct daughter
    nuclide, direct_daughters = get_nuclide_and_direct_progeny(z, a)
    ion_name = nuclide.nuclide
    print(f"Ion : {ion_name} ({z} {a})  ->  direct daughters = {direct_daughters}")

    # units
    nm = g4_units.nm
    m = g4_units.m
    mm = g4_units.mm
    Bq = g4_units.Bq
    kBq = 1000 * Bq

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.number_of_threads = 1
    sim.visu = False
    sim.random_seed = 123456

    # activity
    activity = 10 * kBq / sim.number_of_threads

    # world size
    world = sim.world
    world.size = [10 * m, 10 * m, 10 * m]
    world.material = "G4_WATER"

    # physics
    sim.physics_list_name = "G4EmStandardPhysics_option4"
    sim.physics_manager.enable_decay = True
    sim.physics_manager.global_production_cuts.all = 10 * mm
    sim.physics_manager.global_production_cuts.gamma = 0.001 * mm
    sim.g4_commands_after_init.append("/process/em/pixeXSmodel ECPSSR_ANSTO")
    sim.g4_commands_before_init.append("/process/em/fluoBearden true")

    # sources
    # sim.running_verbose_level = gate.EVENT
    source = sim.add_source("GenericSource", ion_name)
    source.particle = f"ion {z} {a}"
    source.position.type = "sphere"
    source.position.radius = 1 * nm
    source.position.translation = [0, 0, 0]
    source.direction.type = "iso"
    source.activity = activity

    # add stat actor
    s = sim.add_actor("SimulationStatisticsActor", "stats")
    s.track_types_flag = True
    s.output_filename = paths.output / f"test053_{ion_name}_stats.txt"

    # phsp actor
    phsp = sim.add_actor("PhaseSpaceActor", "phsp")
    phsp.attached_to = "world"
    phsp.attributes = [
        "KineticEnergy",
        "EventID",
        # "EventPosition",
        # "EventKineticEnergy",
        # "EventDirection",
        "TrackID",
        "ParentID",
        "TrackCreatorProcess",
        "TrackCreatorModelIndex",
        "ParticleName",
    ]
    phsp.output_filename = paths.output / f"test053_{ion_name}.root"
    phsp.steps_to_store = "exiting first"
    # phsp.store_absorbed_event = True
    # phsp.debug = True

    return ion_name, direct_daughters


def update_sim_for_tac(sim, ion_name, nuclide, activity, end):
    # change simulation parameters
    phsp = sim.get_actor("phsp")

    def rm_type(name, phsp):
        fg = sim.add_filter("ParticleFilter", f"fp_{name}")
        fg.particle = name
        fg.policy = "reject"
        phsp.filters.append(fg)

    phsp.attributes = ["ParticleName", "ParticleType", "GlobalTime"]
    rm_type("gamma", phsp)
    rm_type("anti_nu_e", phsp)
    rm_type("alpha", phsp)
    rm_type("e-", phsp)

    sec = g4_units.second
    Bq = g4_units.Bq

    source = sim.get_source_user_info(ion_name)

    half_life = nuclide.half_life("s") * sec
    lifetime = half_life / math.log(2.0)
    decay_constant = math.log(2.0) / half_life
    print("Ion half life (sec)", half_life / sec)
    print("Ion lifetime (sec)", lifetime / sec)
    print("Ion decay lambda (in s^-1)", decay_constant)

    source.activity = activity
    source.half_life = half_life

    """
    # should work but much too long !
    source.activity = 0
    source.user_particle_life_time = lifetime
    source.n = int((activity / Bq) * (lifetime / sec))"""

    print("Activity  = ", activity / Bq)
    print("Source n  = ", source.n)
    print("Source ac  = ", source.activity / Bq)
    print(f"Source HL = {half_life / sec} sec")
    print(f"Source HL = {half_life / sec / 3600 / 24} days")
    print(f"Source LT = {lifetime / sec} sec")

    # ui = sim.user_info
    # sim.g4_verbose = True
    # sim.g4_commands_after_init.append("/tracking/verbose 2")
    km = g4_units.km
    sim.physics_manager.global_production_cuts.all = 10 * km
    sim.run_timing_intervals = [[0, end]]


def analyse_ion_gamma_from_root(filename, ion_names, events_nb):
    # open file and tree
    root = uproot.open(filename)
    tree = root[root.keys()[0]]
    print(f"Root tree {root.keys()} n={tree.num_entries}")
    print(f"Keys:{tree.keys()}")

    # loop on data to gather all gammas from RadioactiveDecay
    # and their corresponding parent
    i = 0
    event = -1
    g_by_ion = {}
    track = {}
    for batch in tree.iterate(step_size="50MB"):
        # (one single batch)
        print(f"batch len {len(batch)}")
        for e in batch:
            # update current list of track
            if event != e["EventID"]:
                # print(f'New EVENT {e}')
                track = {}
                event = e["EventID"]
            track[e["TrackID"]] = e
            # print(
            #    f'    current list of track {len(track)} -> {e["EventID"]} parent={e["ParentID"]} track={e["TrackID"]} '
            #    f'{e["ParticleName"]} ')

            # sometimes the parent is not in the root file, because this is the track
            # of the primary ion which decay without step
            if e["ParticleName"] == "gamma":
                # if e["TrackCreatorProcess"] == "RadioactiveDecay": # G4 11.1
                # if e["TrackCreatorProcess"] == "Radioactivation": # G4 11.2
                if e["TrackCreatorModelIndex"] == 130:
                    pid = e["ParentID"]
                    if pid not in track:
                        if pid == 1:
                            # it means, the parent is the initial ion (not in the phsp because no step)
                            continue
                        print()
                        print(e)
                        print("event id", e["EventID"])
                        print(f"track {len(track)}")
                        print(f"track {track}")
                        sys.exit(0)
                    ion = track[e["ParentID"]]["ParticleName"]
                    # ene = e["KineticEnergy"]
                    # if ene < 100 * keV:
                    #    print(f"read {e} {ene/keV} keV {ion}")
                    if ion not in g_by_ion:
                        g_by_ion[ion] = []
                    g_by_ion[ion].append(e)
            i += 1

    print(f"Found {len(g_by_ion)} different gamma lines")

    # filter to keep only the one from the asked ion
    ion_names = [a.replace("-", "") for a in ion_names]
    print(ion_names)
    filtered_g_by_ion = {}
    for g in g_by_ion:
        print(f"{g} : n={len(g_by_ion[g])}")
        for ion_name in ion_names:
            if ion_name in g:
                filtered_g_by_ion[g] = g_by_ion[g]
    print(f"Keep {ion_names} only, found {len(filtered_g_by_ion)} lines")

    # extract all gammas and merge according
    # to a given energy precision
    gammas = []
    for g in filtered_g_by_ion:
        x = filtered_g_by_ion[g]
        for i in x:
            e = i["KineticEnergy"]
            e = round(e, 3)  # FIXME ?
            gammas.append(e)
    print(f"Found n={len(gammas)} different energy gammas, grouped by 1e-3 precision")

    # histogram
    gamma_peaks = {}
    for g in gammas:
        if g in gamma_peaks:
            gamma_peaks[g] += 1
        else:
            gamma_peaks[g] = 1

    # sort
    gamma_peaks = dict(sorted(gamma_peaks.items()))
    gp_ene = np.array([g for g in gamma_peaks])
    gp_w = np.array([g for g in gamma_peaks.values()])
    gp_w = gp_w / events_nb

    # print
    keV = g4_units.keV
    for e, w in zip(gp_ene, gp_w):
        print(f"{e / keV:.4f} keV \t -> {w * 100:.4f} %")

    return gp_ene, gp_w


def analyse(paths, sim, output, ion_name, z, a, daughters, log_flag=True, tol=0.03):
    # print stats
    stats = output.get_actor("stats")
    print(stats)

    # Monte carlo data
    print("Data from MC, normalized by nb events")
    phsp = sim.get_actor("phsp")
    g2_ene, g2_w = analyse_ion_gamma_from_root(
        phsp.get_output_path(), daughters, stats.counts.events
    )

    # direct computation of gammas
    print()
    print("Data extracted from the database")
    ge = PhotonIonDecayIsomericTransitionExtractor(
        z, a, verbose=True
    )  ## FIXME change verbose
    ge.extract()
    g1_ene = []
    g1_w = []
    keV = g4_units.keV
    for g in ge.gammas:
        print(
            f"{g.transition_energy / keV:.4f} keV \t-> {g.final_intensity * 100:.4f} % "
        )
        g1_ene.append(g.transition_energy)
        g1_w.append(g.final_intensity)
    g1_ene = np.array(g1_ene)
    g1_w = np.array(g1_w)
    print()

    # save to files
    f_mc = str(paths.output / f"test053_{ion_name}_mc.txt")
    with open(f_mc, "w") as f:
        f.write(f"# gamma intensity for {ion_name} -> {' '.join(daughters)}\n")
        f.write("# from Monte Carlo\n")
        for e, w in zip(g2_ene, g2_w):
            f.write(f"{e} {w}\n")

    # save to files
    f_model = str(paths.output / f"test053_{ion_name}_model.txt")
    with open(f_model, "w") as f:
        f.write(f"# gamma intensity for {ion_name} -> {' '.join(daughters)}\n")
        f.write("# from model\n")
        for e, w in zip(g1_ene, g1_w):
            f.write(f"{e} {w}\n")

    # plot reference
    f, ax = plt.subplots(1, 1, figsize=(15, 5))
    ax.bar(
        g2_ene,  # "- 2 * keV,
        g2_w,
        width=0.003,
        label=f"Monte Carlo {ion_name} -> {' '.join(daughters)}",
        color="red",
        alpha=0.7,
        log=log_flag,
    )
    # plot model
    ax.bar(
        g1_ene,  # + 2 * keV,
        g1_w,
        width=0.003,
        label=f"Model {ion_name} -> {' '.join(daughters)}",
        color="blue",
        alpha=0.5,
        log=log_flag,
    )
    ax.set_ylabel("Intensity (log)")
    ax.set_xlabel("Energy in keV")  # (slightly offset by +-2 keV for visualisation)")
    if log_flag:
        ax.set_yscale("log")
    ax.legend()

    f = str(paths.output / f"test053_{ion_name}.pdf")
    plt.savefig(f)

    # compare for tests
    is_ok = True
    for e, w in zip(g1_ene, g1_w):
        # only consider weight larger than 0.2%
        if w > 0.002:
            for e2, w2 in zip(g2_ene, g2_w):
                # match energy ?
                if np.fabs(e2 - e) / e < 0.01:
                    d = np.fabs(w2 - w) / w
                    # tol = 0.03
                    if w < 0.02:
                        tol = 0.5
                    ok = d < tol
                    print_test(
                        ok,
                        f"model={e / keV} keV    MC={e2 / keV} keV"
                        f"   {w * 100:.4f}%  {w2 * 100:.4f}%   => {d * 100:.4f}% (tol={tol})",
                    )
                    is_ok = ok and is_ok

    print()
    print(f"Figure in {f}")
    return is_ok


def analyse_time_per_ion_root(sim, end):
    phsp = sim.get_actor("phsp")
    filename = phsp.get_output_path()
    root = uproot.open(filename)
    print(f"Open root file {filename}")
    tree = root[root.keys()[0]]
    print(f"Root tree {root.keys()} n={tree.num_entries}")
    print(f"Keys:{tree.keys()}")

    # group by ion
    sec = g4_units.s
    time_by_ion = {}
    for batch in tree.iterate():
        for e in batch:
            if e["ParticleType"] != "nucleus":
                continue
            n = e["ParticleName"]
            if "[" in n:
                continue
            if n not in time_by_ion:
                time_by_ion[n] = []
            t = e["GlobalTime"] / sec
            if t < end:
                time_by_ion[n].append(t)

    return time_by_ion



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
import os
import subprocess

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "", output_folder="test032")

    is_ok = True

    # test 1 = 10mm and with shell
    f1 = paths.output / "iec_10mm.mhd"
    f2 = paths.output / "iec_source_10mm.mhd"
    path_voxelize_script = os.path.join(
        os.path.dirname(__file__), "..", "..", "bin", "voxelize_iec_phantom.py"
    )
    cmd = (
        f"python {path_voxelize_script} -o {f1} "
        f"-s 10 "
        f"--output_source {f2} "
        f"-a 666 555 444 333 222 111 "
    )
    print(cmd)
    subprocess.call(cmd.split())

    # test 2 = 9mm and without shell
    f3 = paths.output / "iec_9mm.mhd"
    f4 = paths.output / "iec_source_9mm.mhd"
    cmd = (
        f"python {path_voxelize_script} -o {f3} "
        f"-s 9 "
        f"--output_source {f4} "
        f"-a 111 222 333 444 555 666 "
        f"--no_shell "
    )
    print(cmd)
    subprocess.call(cmd.split())

    # compare images
    print("\nCheck difference with ref image")
    is_ok = (
        utility.assert_images(
            paths.output_ref / "iec_10mm.mhd", f1, stats=None, tolerance=0.001
        )
        and is_ok
    )
    is_ok = (
        utility.assert_images(
            paths.output_ref / "iec_source_10mm.mhd", f2, stats=None, tolerance=0.001
        )
        and is_ok
    )
    is_ok = (
        utility.assert_images(
            paths.output_ref / "iec_9mm.mhd", f3, stats=None, tolerance=0.001
        )
        and is_ok
    )
    is_ok = (
        utility.assert_images(
            paths.output_ref / "iec_source_9mm.mhd", f4, stats=None, tolerance=0.001
        )
        and is_ok
    )
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
import numpy as np
import meshio
import itk


def create_test_mesh(file_name="myTesselatedBoxVolume.stl"):
    mesh = create_box_mesh()
    mesh.write(file_name)


def create_box_mesh():
    """Create a simple cube mesh of 300×300×300 mm³"""

    # Define the vertices and faces of the box using vectors
    vertices = np.array(
        [
            [0, 0, 0],  # Vertex 0
            [1, 0, 0],  # Vertex 1
            [1, 1, 0],  # Vertex 2
            [0, 1, 0],  # Vertex 3
            [0, 0, 1],  # Vertex 4
            [1, 0, 1],  # Vertex 5
            [1, 1, 1],  # Vertex 6
            [0, 1, 1],  # Vertex 7
        ],
        dtype=np.float32,
    )

    # scale vertices from a 1×1×1 mm³ cube to a 0-centered 300×300×300 mm³ cube
    vertices = (vertices - 0.5) * 300

    # Define the faces of the box using vertex indices
    triangles = [
        [0, 3, 1],
        [1, 3, 2],
        [0, 4, 7],
        [0, 7, 3],
        [4, 5, 6],
        [4, 6, 7],
        [5, 1, 2],
        [5, 2, 6],
        [2, 3, 6],
        [3, 7, 6],
        [0, 1, 5],
        [0, 5, 4],
    ]

    cells = [("triangle", triangles)]

    # Create the mesh
    return meshio.Mesh(vertices, cells)


def create_simulation():
    output_path = paths.output

    # units
    m = gate.g4_units.m
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    eV = gate.g4_units.eV
    MeV = gate.g4_units.MeV
    um = gate.g4_units.um

    # simulation object
    sim = gate.Simulation()

    #
    # Visualization
    sim.visu = False
    sim.visu_verbose = False
    sim.random_engine = "MersenneTwister"
    sim.random_seed = 123456
    sim.number_of_threads = 1
    sim.output_dir = output_path

    # geometry
    world = sim.world
    world.size = [2 * m, 2 * m, 2 * m]
    world.material = "G4_Galactic"

    tes = sim.add_volume("Tesselated", name="MyTesselatedVolume")
    tes.material = "G4_WATER"
    tes.file_name = output_path / "myTesselatedBoxVolume.stl"
    tes.origin_at_cog = True

    # print the list of available volumes types:
    sim.volume_manager.print_volume_types()

    # sources
    source = sim.add_source("GenericSource", "particle")
    source.particle = "proton"
    source.position.type = "point"
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.position.translation = [0 * cm, 0 * cm, -30 * cm]
    source.energy.type = "gauss"
    source.energy.mono = 60 * MeV
    source.n = 100

    print(f"The energy is {source.energy.mono / eV} eV")

    # Physics
    sim.physics_manager.physics_list_name = "QGSP_BIC_EMZ"
    global_cut = 500 * um
    sim.physics_manager.global_production_cuts.gamma = global_cut
    sim.physics_manager.global_production_cuts.electron = global_cut
    sim.physics_manager.global_production_cuts.positron = global_cut
    sim.physics_manager.global_production_cuts.proton = global_cut

    # Actors
    # Statistics Actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.output_filename = "Statistics.txt"
    stats.track_types_flag = True

    # Dose Actor
    dose_actor = sim.add_actor("DoseActor", "dose")
    dose_actor.attached_to = tes
    # number of voxels per dimension
    dose_actor.size = [1, 1, 300]
    # size of the voxels
    dose_actor.spacing = [30 * cm, 30 * cm, 1 * mm]
    dose_actor.dose.active = True
    dose_actor.edep_squared.active = True
    dose_actor.hit_type = "random"
    dose_actor.output_filename = "test067_dose.mhd"

    return sim


def eval_results(sim):
    # access to the results
    eval_Volume = sim.volume_manager.get_volume(
        "MyTesselatedVolume"
    ).solid_info.cubic_volume
    print("volume: ", eval_Volume)
    volume_is_ok = utility.check_diff_abs(
        float(eval_Volume), float(27000000.0), tolerance=1e-1, txt="volume"
    )

    dose = sim.get_actor("dose")
    image = dose.edep.get_data()
    np_image = itk.GetArrayFromImage(image)
    # For 1D images, the array is squeezed
    np_image = np.squeeze(np_image)
    # create index array
    array_1d = np.arange(300)
    # Check if the lengths of the arrays match
    if len(array_1d) != len(np_image):
        raise ValueError("Lengths of the arrays do not match")
    (
        r80,
        tmp,
    ) = utility.getRange(array_1d, np_image, percentLevel=0.8)

    print("r80: ", r80)
    r80_is_ok = utility.check_diff_abs(
        float(r80), float(30.43), tolerance=1e-1, txt="R80"
    )

    if volume_is_ok and r80_is_ok:
        is_ok = True
    else:
        is_ok = False
    return is_ok


paths = utility.get_default_test_paths(
    __file__, "test067_tesselated_volume", output_folder="test067"
)


if __name__ == "__main__":
    print("Generating STL data")
    create_test_mesh(file_name=str(paths.output / "myTesselatedBoxVolume.stl"))

    sim = create_simulation()

    print("Running GATE simulation")
    sim.run()

    stats_actor = sim.actor_manager.get_actor("Stats")
    print(stats_actor)

    print("Simulation finished")
    print("Evaluating results")
    is_ok = eval_results(sim)
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
from test013_phys_lists_helpers import create_pl_sim

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "gate_test013_phys_lists")

    # create simulation via the helper function
    sim = create_pl_sim()
    sim.g4_verbose = True

    # remove ion sources
    sim.source_manager.sources.pop("ion1")
    sim.source_manager.sources.pop("ion2")

    # change physics
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option4"
    # enable decay via switch:
    # sim.physics_manager.enable_decay = True
    # or by activating the physics constructors:
    sim.physics_manager.special_physics_constructors.G4DecayPhysics = True
    sim.physics_manager.special_physics_constructors.G4RadioactiveDecayPhysics = True

    um = gate.g4_units.um
    global_cut = 7 * um
    sim.physics_manager.global_production_cuts.gamma = global_cut
    sim.physics_manager.global_production_cuts.electron = global_cut
    sim.physics_manager.global_production_cuts.positron = global_cut
    sim.physics_manager.global_production_cuts.proton = global_cut

    # print cuts
    print("Phys list cuts:")
    print(sim.physics_manager.dump_production_cuts())

    # start simulation
    # sim.g4_commands_after_init.append("/tracking/verbose 1")
    sim.run()

    # Gate mac/main_1.mac
    stats = sim.get_actor("Stats")
    stats_ref = utility.read_stat_file(paths.gate_output / "stat_1.txt")
    is_ok = utility.assert_stats(stats, stats_ref, tolerance=0.12)

    utility.test_ok(is_ok)



from anytree import LoopError
from pathlib import Path
import tables
from collections import OrderedDict
import inspect

import opengate.postprocessors

# from .datafetchers import DataFetcherHdf5
# from .image import ProjectionListMode
# from .listmode import ListModeSingleAttribute, GaussianBlurringSingleAttribute, \
#     OffsetSingleAttribute
# from .sequences import ProcessingSequence
from ..base import GateObject
from ..exception import fatal
from ..utility import ensure_directory_exists

__run_group_prefix__ = "data_from_run"


class ProcessingGroupBase(GateObject):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.processing_units = OrderedDict()
        self._need_tree_update = True  # flag to store state of processing unit tree

    def close(self):
        for r in self.tree_roots:
            r.close()
        self._need_tree_update = True

    def to_dictionary(self):
        d = super().to_dictionary()
        d["processing_units"] = dict(
            [(k, v.to_dictionary()) for k, v in self.processing_units.items()]
        )
        return d

    @property
    def last_added_processing_unit(self):
        return self.processing_units[next(reversed(self.processing_units))]

    def add_processing_unit(self, processing_unit, **kwargs):
        try:
            is_unit = processing_unit.is_processing_unit
        except AttributeError:
            is_unit = False
        if is_unit:
            _processing_unit = processing_unit
        else:
            try:
                processing_unit_class = (
                    opengate.postprocessors.available_processing_units[processing_unit]
                )
            except KeyError:
                fatal(
                    f"Invalid input for processing_unit. "
                    f"Should be an instance of a processing unit class or a known processing unit name. "
                    f"Available classes are: {list(opengate.postprocessors.available_processing_units.keys())}"
                )
            _processing_unit = processing_unit_class(**kwargs)
        if _processing_unit.name not in self.processing_units:
            _processing_unit.processing_group = self
            _processing_unit.simulation = self.simulation
            # automatically set hierarchy, (can be changed manually by user):
            if len(self.processing_units) > 0:
                self.last_added_processing_unit.yields_to = _processing_unit
            if hasattr(_processing_unit, "processing_units"):
                for u in _processing_unit.processing_units.values():
                    u.simulation = self.simulation
            self.processing_units[_processing_unit.name] = _processing_unit
        else:
            fatal(
                f"ProcessingGroup {self.name} already contains a post-processing unit named {_processing_unit.name}. "
            )
        self._need_tree_update = True
        # if the instance has only been created here, return it.
        if _processing_unit is not processing_unit:
            return _processing_unit

    def get_processing_unit(self, name):
        try:
            return self.processing_units[name]
        except KeyError:
            fatal(
                f"No unit with the name {name} found. Known units are: {list(self.processing_units.keys())}"
            )

    @property
    def all_processing_units(self):
        all_units = set()
        for u in self.processing_units.values():
            all_units.add(u)
            try:
                all_units.update(u.all_processing_units)
            except AttributeError:
                pass
        return all_units

    @property
    def is_post_processor(self):
        if isinstance(self, PostProcessor):
            return True
        else:
            return False

    @property
    def tree_roots(self):
        self.update_processing_tree_if_needed()
        tree_roots = set()
        for v in self.processing_units.values():
            tree_roots.add(v.root)
        return tuple(tree_roots)

    def get_output_groups(self):
        return [tr.hdf5_group for tr in self.tree_roots]

    def update_processing_tree(self):
        for v in self.processing_units.values():
            try:
                v.update_tree_node()
            except LoopError:
                fatal(
                    f"There seems to be a loop in the processing unit tree involving unit {v.name}."
                )
            try:
                v.update_processing_tree()
            except AttributeError:
                pass
        self._need_tree_update = False

    def update_processing_tree_if_needed(self):
        if self._need_tree_update:
            self.update_processing_tree()

    @property
    def processing_groups(self):
        return [
            u
            for u in self.processing_units.values()
            if isinstance(u, ProcessingGroupBase)
        ]


class PostProcessor(ProcessingGroupBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.output_file_handle = None
        self.extra_file_handles = {}

    def initialize(self):
        self.update_processing_tree()
        for n in self.all_initial_units:
            # make sure all input units are data fetcher
            if n.can_be_initial_unit is False:
                fatal(f"Input node {n.name} is not a data fetcher. ")

        self.initialize_output_file()
        self.initialize_hdf5_groups()

    def initialize_output_file(self):
        if self.output_file_handle is None:
            self.output_file_handle = tables.open_file(
                self.simulation.get_output_path(f"output_{self.name}.h5"),
                mode="w",
                title=f"Output post-processor {self.name}",
            )
        ensure_directory_exists(self.output_directory_external_files)

    @property
    def output_directory_external_files(self):
        output_directory = self.output_directory
        if output_directory is None:
            return None
        else:
            return output_directory / (
                Path(self.output_file_handle.filename).stem.rstrip("_")
                + "_external_files"
            )

    @property
    def output_directory(self):
        if self.output_file_handle is not None:
            return Path(self.output_file_handle.filename).parent
        else:
            return None

    def get_extra_file_handle(self, identifier):
        try:
            return self.extra_file_handles[str(identifier)]
        except KeyError:
            raise KeyError(
                f"Cannot find an input file handle for identifier '{identifier}'."
            )

    def get_or_register_extra_file_handle(self, file_handle):
        """Register a handle if it is not known yet."""
        identifier = str(file_handle.filename)
        print(f"register_extra_file_handle {identifier}")
        try:
            return self.get_extra_file_handle(identifier)
        except KeyError:
            self.extra_file_handles[identifier] = file_handle
            return file_handle

    def has_extra_file_handle(self, identifier):
        try:
            self.get_extra_file_handle(identifier)
            return True
        except KeyError:
            return False

    # def get_external_link_file_handle(self, identifier):
    #     try:
    #         return self.external_link_file_handles[identifier]
    #     except KeyError:
    #         raise KeyError(f"Cannot find an external link file handle for identifier '{identifier}'.")
    #
    # def register_external_link_file_handle(self, file_handle):
    #     """Register a handle if it is not known yet. """
    #     identifier = file_handle.filename
    #     if self.get_extra_file_handle(identifier) is None:
    #         self.external_link_file_handles[identifier] = file_handle

    def create_external_link(self, where, name, target, createparents=False):
        """Create an external link.

        Create an external link to a *target* node with the given *name*
        in *where* location.  *target* can be a node object in another
        file or a path string in the form 'file:/path/to/node'.  If
        *createparents* is true, the intermediate groups required for
        reaching *where* are created (the default is not doing so).

        The returned node is an :class:`ExternalLink` instance.

        """

        if not isinstance(target, str):
            if hasattr(target, "_v_pathname"):  # quacks like a Node
                target = target._v_file.filename + ":" + target._v_pathname
            else:
                raise ValueError("`target` has to be a string or a node object")
        elif target.find(":/") == -1:
            raise ValueError("`target` must expressed as 'file:/path/to/node'")
        parentnode = self.output_file_handle._get_or_create_path(where, createparents)
        elink = ManagedExternalLink(self, parentnode, name, target)
        # Refresh children names in link's parent node
        parentnode._g_add_children_names()
        return elink

    def release_file_handles(self):
        for k, v in self.extra_file_handles.items():
            v.close()
        self.extra_file_handles = {}
        self.output_file_handle.close()
        self.output_file_handle = None

    def close(self):
        super().close()  # this closes all data handles
        self.release_file_handles()

    def run(self):
        try:
            self.initialize()
            for tr in self.tree_roots:
                tr.update_output()
            self.output_file_handle.flush()
        finally:
            self.close()

    @property
    def hdf5_group(self):
        if self.output_file_handle is not None:
            return self.output_file_handle.get_node("/")
        else:
            return None

    def initialize_hdf5_groups(self):
        for u in self.all_processing_units:
            u.initialize_hdf5_group_in_output_file()

    @property
    def all_initial_units(self):
        """Get all"""
        initial_units = []
        for tr in self.tree_roots:
            initial_units.extend(tr.leaves)
        return initial_units


class ManagedExternalLink(tables.linkextension.ExternalLink, tables.link.Link):
    """Variant of the ExternalLink class from PyTables."""

    # Class identifier.
    _c_classid = "MANAGEDEXTERNALLINK"

    def __init__(self, post_processor, parentnode, name, target=None, _log=False):
        # self.extfile = None  # PyTables' ExternalLink class defines this attribute.
        self.post_processor = post_processor
        """The external file handler, if the link has been dereferenced.
        In case the link has not been dereferenced yet, its value is
        None."""
        super().__init__(parentnode, name, target, _log)

    @property
    def extfile(self):
        filename, target = self._get_filename_node()
        absolute_path = str(Path(filename).absolute())
        try:
            return self.post_processor.get_extra_file_handle(absolute_path)
        except KeyError:
            return None

    @extfile.setter
    def extfile(self, file_handle):
        self.post_processor.get_or_register_extra_file_handle(file_handle)

    def _get_filename_node(self):
        """Return the external filename and nodepath from `self.target`."""

        # This is needed for avoiding the 'C:\\file.h5' filepath notation
        filename, target = self.target.split(":/")
        return filename, "/" + target

    def __call__(self, **kwargs):
        """Dereference self.target and return the object.

        You can pass all the arguments supported by the :func:`open_file`
        function (except filename, of course) so as to open the referenced
        external file.
        """

        filename, target = self._get_filename_node()

        if not Path(filename).is_absolute():
            # Resolve the external link with respect to the this
            # file's directory.  See #306.
            filename = str(Path(self._v_file.filename).with_name(filename))

        if self.extfile is None or not self.extfile.isopen:
            self.extfile = tables.open_file(filename, **kwargs)
        else:
            # XXX: implement better consistency checks
            assert self.extfile.filename == filename
            # assert self.extfile.mode == kwargs.get('mode', 'r')

        return self.extfile._get_node(target)

    def umount(self):
        """Safely unmount self.extfile, if opened."""

        extfile = self.extfile
        # Close external file, if open
        if extfile is not None and extfile.isopen:
            extfile.close()
            self.extfile = None

    # def _f_close(self):
    #     """Especific close for external links."""
    #
    #     self.umount()
    #     super()._f_close()

    def __str__(self):
        """Return a short string representation of the link."""

        return f"{self._v_pathname} ({self.__class__.__name__}) -> " f"{self.target}"


# available_processing_units = {
#     'ProcessingSequence': ProcessingSequence,
#     'DataFetcherHdf5': DataFetcherHdf5,
#     'ProcessingUnitListModeSingleAttribute': ListModeSingleAttribute,
#     'GaussianBlurringSingleAttribute': GaussianBlurringSingleAttribute,
#     'OffsetSingleAttribute': OffsetSingleAttribute,
#     'ProjectionFromListMode': ProjectionListMode
# }



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from test061_user_event_info_helpers import create_simulation, analyse
from opengate.tests import utility


if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "", output_folder="test061")

    # create the simulation
    sim = gate.Simulation()
    sim.number_of_threads = 1
    create_simulation(sim, paths, "mono")

    # run
    sim.run(start_new_process=True)

    # analyse 1
    is_ok = analyse(sim)

    # run in MT
    sim.number_of_threads = 2
    sim.run(start_new_process=True)

    # analyse 2
    is_ok = analyse(sim) and is_ok

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import opengate as gate
import test049_pet_digit_blurring_helpers as t49
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "gate_test049_pet_blur", "test049")

    """
    see https://github.com/teaghan/PET_MonteCarlo
    and https://doi.org/10.1002/mp.16032

    PET simulation to test blurring options of the digitizer

    - PET:
    - phantom: nema necr
    - output: singles with and without various blur options
    """

    # create the simulation
    sim = gate.Simulation()
    t49.create_simulation(sim, singles_name="Singles_readout", fname_suffix="_v3")

    # const
    ns = gate.g4_units.ns
    keV = gate.g4_units.keV
    sigma_to_fwhm = 2 * np.sqrt(2 * np.log(2))
    fwhm_to_sigma = 1.0 / sigma_to_fwhm

    # add (fake) blur
    ro = sim.get_actor("Singles_readout")
    bc1 = sim.add_actor("DigitizerBlurringActor", "Singles_1")
    bc1.output_filename = ro.output_filename
    bc1.input_digi_collection = "Singles_readout"
    bc1.blur_attribute = "GlobalTime"
    bc1.blur_method = "Gaussian"
    bc1.blur_fwhm = 100 * ns

    bc2 = sim.add_actor("DigitizerBlurringActor", "Singles")
    bc2.output_filename = ro.output_filename
    bc2.input_digi_collection = bc1.name
    bc2.blur_attribute = "TotalEnergyDeposit"
    bc2.blur_method = "InverseSquare"
    bc2.blur_resolution = 0.18
    bc2.blur_reference_value = 511 * keV

    # start simulation
    sim.run()

    # print results
    stats = sim.get_actor("Stats")
    print(stats)

    # ----------------------------------------------------------------------------------------------------------
    readout = sim.get_actor("Singles_readout")
    ig = readout.GetIgnoredHitsCount()
    print()
    print(f"Nb of ignored hits : {ig}")

    # check stats
    print()
    gate.exception.warning(f"Check stats")
    p = paths.gate_output
    stats_ref = utility.read_stat_file(p / "stats_blur.txt")
    is_ok = utility.assert_stats(stats, stats_ref, 0.025)

    # check root singles
    f = p / "pet_blur.root"
    bc = sim.get_actor("Singles")
    is_ok = (
        t49.check_root_singles(
            paths, 1, f, bc.get_output_path(), png_output="test049_singles_wb.png"
        )
        and is_ok
    )

    # timing
    b = t49.check_timing(f, bc.get_output_path())
    is_ok = is_ok and b

    utility.test_ok(is_ok)
